/* -----------------     7SEG_TEMP.C    -------------------

     2-stelliges Thermometer mit Anzeige auf einem 7-Segment
     LCD.
     
     "Messschaltung":   Spannungsteiler mit 10k Ohm Fest- 
                        widerstand und 10k Ohm NTC gegen
                        Masse. Teilerabgriff an PIN ADC0

     Verwendbare Controller:

     ATmega8 / ATmega48 .. ATmega328

     Slave-Adresse:    0x40 (siehe Defines)
     verwendete MCU:   ATmega8


     Quarz:            keiner / int. 8MHz Takt
     Fuses             L-Fuse:  0xD4
                       H-Fuse:  0xC9

     Compiler:         AVR-GCC  4.7.2
     eingebunden:      lprintf_flt (automatisch mit Batch-Datei)

     IDE:              Context

     uebersetzen mit:  GCCCOMPILE_GKZ.BAT

     Originalausgangspunkt: Atmel Note AVR241

     modifiziert und erweitert:
     20.06.2013        R. Seelig

   ------------------------------------------------------- */

/*
     _a_            Pinzuordnung DE-112 Series (bspw. Reichelt Best.Nr.: LCD 2,0-13
    |   |
   f|   |b              Display                 Controller
     -g-            Funktion   PIN            PIN    Funktion
    |   |
   e|___|c             BP       1 ------------ 10      PB7*
      d                DP1      2 ------------ nicht angeschlossen
                       1e       3 ------------  6      PD4
                       1d       4 ------------  5      PD3
                       1c       5 ------------  4      PD2
                       DP2      6 ------------ 13      PD7
                       2e       7 ------------ 18      PB4
                       2d       8 ------------ 17      PB3
                       2c       9 ------------ 16      PB2

                       2b      10 ------------ 15      PB1
                       2a      11 ------------ 14      PB0
                       2f      12 ------------ 19      PB5
                       2g      13 ------------  9      PB6
                       1b      14 ------------  3      PD1
                       1a      15 ------------  2      PD0
                       1f      16 ------------ 11      PD5
                       1g      17 ------------ 12      PD6
                       BP      18 ------------ 10      PB7*

                       * beide Backplanes (BP) sind miteinander verbunden

   -------------------------------------------------------
   
   Spannungsteiler:
   
                   o +5V
                   |
                   _
                  | |  10k
                  |_|
                   |
                   o----o ADC0
                   |
                   _
                  | |  10k NTC
                  |_|
                   |
                  ---
*/


#define F_CPU 8000000
#define QUARZ F_CPU
#define t1reload 200

#include <util\delay.h>
#include <avr/io.h>
#include <avr\interrupt.h>


// ---------------------- Timer  -------------------------

#ifndef TIMSK1
  #define TIMSK1 TIMSK                             // beim ATmega8 gibt es kein TIMSK1, deshalb wird TIMSK "umbenannt"
#endif

// ----------------------- ADC ---------------------------

#define adc_low        ADCL
#define adc_high       ADCH


// --------------- NTC-Materialkonstanten ----------------

//  Definitionen fuer NTC-02 - 47k lt. Datenblatt   ( NTC-02 47k von Reichelt)

/*
#define ma          3.354016e-3
#define mb          2.51911e-4
#define mc          3.51094e-12
#define r25         47

*/

//  Definitionen fuer NTC-02 - 10k lt. Datenblatt   ( NTC-02 10k von Reichelt)


#define ma          3.354016e-3
#define mb          2.569850e-4
#define mc          2.620130e-12
#define r25         10

/* -----------------------------------------------
         7-Segment-Anzeige Routinen
   ----------------------------------------------- */

// Bitmuster fuer Segmente und den Dot-Punkt

uint8_t seg7table [] = {0x3f,0x06,0x5b,0x4f,0x66,0x6d,0x7d,0x07,0x7f,0x6f,0x77,0x7c,0x39,0x5e,0x79,0x71};

uint8_t volatile dotpoint;
uint8_t volatile timcount;
uint8_t lcd_digit1, lcd_digit2;


void lcd_update(void)
{
  static char sequence=0;
  if(!sequence)
  {
    PORTB = lcd_digit1 &= ~0x80;
    if (dotpoint==0) {PORTD = lcd_digit2;} else {PORTD =lcd_digit2 | 0x80;}
    sequence = 1;
  }
  else
  {
    PORTD = PORTD ^ 0xff;
    PORTB = PORTB ^ 0xff;
    sequence=0;
  }
}

ISR (TIMER1_COMPA_vect)
{
  if( --timcount == 0 )
  {
    timcount = 5;
    lcd_update();
  }
}

void lcd_print(char digit, char lcdnum)
{
  if(digit == 1)
  {
    lcd_digit1 = seg7table[lcdnum];             //Updates global register LCD.digit1
  }
  else
  {
    lcd_digit2 = seg7table[lcdnum];             //Updates global register LCD.digit2
  }
}

void print7seg_hex(uint8_t wert)
{
  lcd_print(2,wert>>4);
  lcd_print(1,wert& 0x0f);
}

void print7seg_dez(uint8_t wert)
{
  lcd_print(2,wert / 10);
  lcd_print(1,wert % 10);
}

//        Ende 7-Segment-Implementation
/* ----------------------------------------------- */

/* -----------------------------------------------
        Reloadwerte fuer Timerinterrupt
   ----------------------------------------------- */


void init_timer1(void)
{
  TCCR1B = 1<<WGM12^1<<CS10;
  OCR1A = QUARZ / t1reload-1;
  TCNT1 = 0;
  timcount= 5;

  TIMSK1 = 1<<OCIE1A;
  sei();
}

/* -----------------------------------------------
         ADC - Routinen
   ----------------------------------------------- */

unsigned int getadc_10bit (void)
{
  unsigned int adcwert,adch;

  adcwert= adc_low;
  adch= adc_high;
  adch = adch<<8;
  adcwert|= adch;
  return(adcwert);
}

void adc_init(uint8_t refmode, uint8_t channel)
{
/* <refmode> bezeichnet die zu verwendende Referenz
     0 : externe Referenz an Aref
     1 : +AVcc mit ext. Kondensator
     2 : reserviert
     3 : interne Referenz: 1,1V ATmega168 / 2,56V ATmega8

   <channel> bezeichnet den Kanal auf dem gemessen werden soll

Bit: 7      6
   REFS1  REFS0      Referenz
     0      0        ext. Referenz an  Aref
     0      1        +AVcc mit ext. Kondensator
     1      0        reserviert
     1      1        Interne Spannungsreferenz: 1,1V (ATmega168) / 2,56V (ATmega8)

  Bit5: ADLAR = 0 (rechtsbuendig); MUX= 0 => ADC0 gewaehlt

Bit3..0: Selektor Analogmultiplexer


   // ADMUX= 1<<REFS0;                                          // AVcc (Betriebsspannung) als Referenz
   // ADMUX= 1<<REFS0 | 1<<REFS1;                               // int. Referenz
*/



   ADMUX = refmode<<6;
   ADMUX |= channel;

   ADCSRA= 0xe7;						// ADC enable, single gestartet, free running mode
           							// Taktteiler / 128
   _delay_ms(1);
   ADCSRA= 0xa7;
}

//        Ende ADC Implementation
/* ----------------------------------------------- */

/* ----------------------------------------------------------------------------------------------------------
                                                    M A I N
   ---------------------------------------------------------------------------------------------------------- */

int main(void)
{
  float     rmess;
  float     rl;
  float     za,zb,zc;

  uint16_t  mwert;
  uint16_t  addmess;
  
  int b;


  DDRB = 0xff;                                  // Pins, an denen das 7-Segmentdisplay angeschlossen ist auf
  DDRD = 0xff;                                  // Ausgang schalten

  init_timer1();                                // Timing fuer die Interruptgesteuerten Wechsel des Displays
  dotpoint= 0;                                  // Kommapunkt am Display aus

  adc_init(1,0);
  while(1)
  {
    addmess= 0;
    for (b= 0; b<10; b++)                             // 10 mal erfassen
    {
      addmess += getadc_10bit();
      _delay_ms(30);
    }

    mwert= addmess/10;                                // und Mittelwert bilden


                                                      // Spannungsteiler mit 10k Vorwiderstand an Betriebsspannung
                                                      // Spannungsteilerformel: (R2/Ur2) = (R1/Ur1)
                                                      //                         R2 = (R1/Ur1)*Ur2
    rmess= (float)((float)10.0/(1024-mwert))*mwert;   // da nur Verhaeltnis von Spannung zu Widerstaenden eingeht, ist
                                                      // die absolute Genauigkeit der Spannung untergeordnet, die
                                                      // Versorgungsspannung des Spannungsteilers betraegt 1024 Digits

    za= ma;                                           // Naeherung des Temperaturwertes nach Steinhart - Hart
    rl= log(rmess/r25);
    zb= mb*rl;
    rl= rl*rl;
    zc= mc*rl;
    rl= (1/(za+zb+zc))-273.3;                         // Temperaturergebnis wird in K ermittelt, deshalb ist
    
    b= rl;                                            // dieser Wert (273,3) abzuziehen, rl beinhaltet Wert in C
    print7seg_dez(b);
  }
}

