/* -----------------     I2C_7SEG.C    -------------------

     Generische 2-stellige LCD 7-Segmentanzeige. Ansteuer-
     ung ueber den I2C Bus.
     Ohne I2C - Ansteuerung zeigt das Display seine
     Slave-Adresse hexadezimal an. Die Anzeige ueber den
     I2C Bus erfolgt dezimal (siehe Defines).


     Verwendbare Controller:

     ATmega8 / ATmega48 .. ATmega328

     Slave-Adresse:    0x40 (siehe Defines)
     verwendete MCU:   ATmega8


     Quarz:            keiner / int. 8MHz Takt
     Fuses             L-Fuse:  0xD4
                       H-Fuse:  0xC9

     Compiler:         AVR-GCC  4.7.2
     IDE:              Context

     uebersetzen mit:  GCCCOMPILE.BAT

     Originalausgangspunkt: Atmel Note AVR241

     modifiziert und erweitert:
     20.06.2013        R. Seelig

   ------------------------------------------------------- */

/*
     _a_            Pinzuordnung DE-112 Series (bspw. Reichelt Best.Nr.: LCD 2,0-13
    |   |
   f|   |b              Display                 Controller
     -g-            Funktion   PIN            PIN    Funktion
    |   |
   e|___|c            *BP       1 ------------ 10      PB7
      d                DP1      2 ------------ nicht angeschlossen
                       1e       3 ------------  6      PD4
                       1d       4 ------------  5      PD3
                       1c       5 ------------  4      PD2
                       DP2      6 ------------ 13      PD7
                       2e       7 ------------ 18      PB4
                       2d       8 ------------ 17      PB3
                       2c       9 ------------ 16      PB2

                       2b      10 ------------ 15      PB1
                       2a      11 ------------ 14      PB0
                       2f      12 ------------ 19      PB5
                       2g      13 ------------  9      PB6
                       1b      14 ------------  3      PD1
                       1a      15 ------------  2      PD0
                       1f      16 ------------ 11      PD5
                       1g      17 ------------ 12      PD6
                      *BP      18 ------------ 10      PB7

                       * beide Backplanes (BP) sind mit einander verbunden angeschlossen


                                                               I2C-Bus                     A1   A0    Slaveadresse
                                               28      SCL   ------------                   0    0        0x40
                                               27      SDA   ------------                   0    1        0x42
                                               26      A1    Slave-Adresse-Offset           1    0        0x44
                                               25      A0    Slave-Adresse-Offset           1    1        0x46
*/

#define F_CPU 8000000
#define QUARZ F_CPU

#include <util\delay.h>
#include <avr/io.h>
#include <avr\interrupt.h>

#include "I2C_slave.inc"

#define t1reload 200                            // Taktgeschwindigkeit, mit der die Spannung der Segmente
                                                // gewechselt werden. Kleinerer Wert = schnellere Wechsel


// --------------------------------------------------------------------------------------------------------------

#define slv_addr            0x40                // Grundadresse unter auf die die Schaltung im Bus "hoert"
                                                // Niederwertigstes Bit wird wegen R/W nicht ausgewertet und
                                                // muss 0 bleiben, d.h. es sind nur geradzahlige Adressen zu-
                                                // laessig

#define dez_anzeige                             // Anzeige erfolgt dezimal. Wird dieser Define auskommentiert,
                                                // so erfolgt die Anzeige hexadezimal

// --------------------------------------------------------------------------------------------------------------

                                                // an 2 freien Pins kann bsw. ueber Jumper zusaetzlich
                                                // eine Offsetadresse zur Slaveadresse von aussen gewaehlt
                                                // werden. Somit ist es moeglich, dass insgesamt 4 7-Segment
                                                // Treibereinheiten im Bus vorhanden sind (somit ist es
                                                // mglich, eine insgesamt 8 stellige Anzeige aufzubauen !!!

#define slvddr              DDRC                // Directionregister
#define slvport             PORTC               //
#define slvpins             PINC                // die Eingangspins
#define slvaddr0            PC2
#define slvaddr1            PC3                 // werden beide Pins mit +Ub verdrahtet, erhaehlt der Chip
                                                // die Slaveadresse 0x46 : Adresse = Grundadresse + ((ad0+ad1)*2)

#define isbit(reg,pos)      (reg & (1<<pos))    // Ist ein Bit im Register reg an Position pos gesetzt ?


#ifndef TIMSK1
  #define TIMSK1 TIMSK                          // Timer - Definition fuer ATmega8
#endif

uint8_t i2c_addr;


/* -----------------------------------------------
         7-Segment-Anzeige Routinen
   ----------------------------------------------- */

// Bitmuster fuer Segmente und den Dot-Punkt

uint8_t seg7table [] = {0x3f,0x06,0x5b,0x4f,0x66,0x6d,0x7d,0x07,0x7f,0x6f,0x77,0x7c,0x39,0x5e,0x79,0x71};

uint8_t volatile dotpoint;
uint8_t volatile timcount;
uint8_t lcd_digit1, lcd_digit2;


void lcd_update(void)
{
  static char sequence=0;
  if(!sequence)
  {
    PORTB = lcd_digit1 &= ~0x80;
    if (dotpoint==0) {PORTD = lcd_digit2;} else {PORTD =lcd_digit2 | 0x80;}
    sequence = 1;
  }
  else
  {
    PORTD = PORTD ^ 0xff;
    PORTB = PORTB ^ 0xff;
    sequence=0;
  }
}

ISR (TIMER1_COMPA_vect)
{
  if( --timcount == 0 )
  {
    timcount = 5;
    lcd_update();
  }
}

void lcd_print(char digit, char lcdnum)
{
  if(digit == 1)
  {
    lcd_digit1 = seg7table[lcdnum];             //Updates global register LCD.digit1
  }
  else
  {
    lcd_digit2 = seg7table[lcdnum];             //Updates global register LCD.digit2
  }
}

void print7seg_hex(uint8_t wert)
{
  lcd_print(2,wert>>4);
  lcd_print(1,wert& 0x0f);
}

void print7seg_dez(uint8_t wert)
{
  lcd_print(2,wert / 10);
  lcd_print(1,wert % 10);
}

void init_timer1(void)
{
  TCCR1B = 1<<WGM12^1<<CS10;
  OCR1A = QUARZ / t1reload-1;
  TCNT1 = 0;
  timcount= 5;

  TIMSK1 = 1<<OCIE1A;
  sei();
}

//        Ende 7-Segment-Implementation
/* ----------------------------------------------- */


int main(void)
{

  uint8_t zahl;

  slvddr &= ~(1<<slvaddr0 | 1<<slvaddr1);            // Adresspins auf Eingang schalten
  slvport |= (1<<slvaddr0 | 1<<slvaddr1);            // Pop-Up Widerstaende der Eingaenge

  i2c_addr = slv_addr;

  if (isbit(slvpins,slvaddr0)) i2c_addr += 2;        // Slave-Adresse je nach Aussenbeschaltung
  if (isbit(slvpins,slvaddr1)) i2c_addr += 4;        // festlegen


  DDRB = 0xff;                 // Richtung der Ports, an denen die Segmente des Displays angeschlossen sind
  DDRD = 0xff;

  init_timer1();

  dotpoint= 0;
  print7seg_hex(i2c_addr);     // solange keine neue Zahl auf I2C eintrifft zeigt die Schaltung seine
                               // eigene Slave-Adresse an

  I2C_init(i2c_addr);          // I2C Initialisieren, Slaveadresse 32 (willkuerlich)

  sei();                       // Interrupt fuer den Slave I2C zulassen
  while(1)
  {
    while(rxbuffer[2]==0);     // solange keine Uebernahme eingetroffen ist => warten

    zahl= rxbuffer[0];         // 7 Segmentanzeige bekommt neue Zahl
    if (rxbuffer[1])           // Punkt anzeigen ?
      dotpoint=1;
    else dotpoint=0;

    rxbuffer[2]= 0;            // Uebernahmeanforderung loeschen

#ifdef dez_anzeige
    print7seg_dez(zahl);       // fuer dezimale Darstellung
#else
    print7seg_hex(zahl);       // fuer hexadezimale Darstellung
#endif


  }
}
