/* --------------     I2C_SLAVE.INC   ---------------------

     Routinen fuer einen I2C-Slave
     Unterstuetzt folgende Controller:


     MCU:          ATmega8   / ATmega48  / ATmega88 /
                   ATmega168 / ATmega328

     Compiler:         AVR-GCC
     IDE:              Context

     20.06.2013        R. Seelig

   ------------------------------------------------------- */

#include <util/twi.h>

#define bufferanz    32                       // Puffergroesse fuer einen Vorgang

/* -------------------------------------------------------
     Prototypen
   ------------------------------------------------------- */

volatile uint8_t buffer_address;
volatile uint8_t txbuffer[bufferanz];
volatile uint8_t rxbuffer[bufferanz];

void I2C_init(uint8_t address);
void I2C_stop(void);
ISR(TWI_vect);

/* -------------------------------------------------------
     Implementation
   ------------------------------------------------------- */

void I2C_init(uint8_t address)
{
  TWAR = address;
  TWCR = (1<<TWIE) | (1<<TWEA) | (1<<TWINT) | (1<<TWEN);
}

void I2C_stop(void)
{
  TWCR &= ~( (1<<TWEA) | (1<<TWEN) );
}

ISR(TWI_vect)
{
  uint8_t data;

  if( (TWSR & 0xF8) == TW_SR_SLA_ACK )
  {
    buffer_address = bufferanz;
    TWCR |= (1<<TWIE) | (1<<TWINT) | (1<<TWEA) | (1<<TWEN);
  }
  else if( (TWSR & 0xF8) == TW_SR_DATA_ACK )
  {                                                        // Data ist im "Receive-Mode" eingetroffen
    data = TWDR;                                           // zwischenspeichern

    if(buffer_address == bufferanz)                        // komplett versendet worden ?
    {
      buffer_address = data;
      TWCR |= (1<<TWIE) | (1<<TWINT) | (1<<TWEA) | (1<<TWEN);
    }
    else                                                   // Datenbyte ist komplett
    {
      rxbuffer[buffer_address] = data;                     // im Eingangspuffer ablegen
      buffer_address++;
      if (buffer_address> bufferanz) buffer_address= 0;    // bis zur max. Grenze schreiben, ansonsten Ringpuffer
      if(buffer_address < bufferanz)
      {
        TWCR |= (1<<TWIE) | (1<<TWINT) | (1<<TWEA) | (1<<TWEN);
      }
      else
      {
	TWCR |= (1<<TWIE) | (1<<TWINT) | (0<<TWEA) | (1<<TWEN);
      }
    }
  }
  else if( (TWSR & 0xF8) == TW_ST_DATA_ACK )             // Sendemodus
  {
    data = TWDR;
    if( buffer_address == bufferanz )
    {
      buffer_address = data;
    }
    TWDR = txbuffer[buffer_address];
    buffer_address++;
    if (buffer_address> bufferanz) buffer_address= 0;
    if(buffer_address < bufferanz)
    {
      TWCR |= (1<<TWIE) | (1<<TWINT) | (1<<TWEA) | (1<<TWEN);
    }
    else
    {
      TWCR |= (1<<TWIE) | (1<<TWINT) | (0<<TWEA) | (1<<TWEN);
    }
  }
  else
  {
    TWCR |= (1<<TWIE) | (1<<TWEA) | (1<<TWEN);
  }
}

/* -------------------------------------------------------
     Ende Implementation
   ------------------------------------------------------- */
