/* -------------------     TX_LCD.INC    ------------------

     grundlegende Funktionen zum Betrieb eines
     LC-Textdisplays an einem AVR-Microcontroller.

     Hier am ATtiny2313

     13.06.2013        R. Seelig

   ------------------------------------------------------- */
   
/*
      Anschluss Pollin-Display C0802-04 an den Controller
      ---------------------------------------------------
         o +5V
         |                            Display                 Controller
         _                        Funktion   PIN            PIN    Funktion
        | |
        |_|                          GND      1 ------------
         |                          +5V       2 ------------
         o----o Kontrast   ---    Kontrast    3 ------------
        _|_                           RS      4 ------------  9      PD5
        \ /                          GND      5 ------------
        -'-                    (Takt) E       6 ------------  8      PD4
         |                           D4       7 ------------  2      PD0
        --- GND                      D5       8 ------------  3      PD1
                                     D6       9 ------------  6      PD2
                                     D7      10 ------------  7      PD3
*/                                        

#include <util/delay.h>
#include <avr/io.h>
#include <avr/pgmspace.h>



/* -------------------------------------------------------
     Macros der Pindefinitionen, an welche Pins das
     Textdisplay angeschlossen wird
   ------------------------------------------------------- */


#define lcd_dbio         DDRD				  // Richtungsregister der Datenbits
#define lcd_db           PORTD				  // Datenleitungen sind an PortB angeschlossen

#define lcd_d7           PD3
#define lcd_d6           PD2
#define lcd_d5           PD1
#define lcd_d4           PD0


#define lcd_ctrlio       DDRD				  // Richtungsregister der Steuerbits
#define lcd_ctrl         PORTD				  // Steuerleitungen (RS und E) sind an PortB angeschlossen

#define lcd_rs           PD5			          // RS Leitung Display auf PB4 anschliessen
#define lcd_e            PD4				  // E Leitung Display auf PB5 anschliessen

#define prints(tx)       (putromstring(PSTR(tx)))         // Anzeige String aus Flashrom: prints("Text");

/* -------------------------------------------------------
     diverse Macros
   ------------------------------------------------------- */

#define setbit(reg,pos)  (reg|= 1<<pos)                   // Setzt Bit in reg (Port oder Variable) an Position pos
#define clrbit(reg,pos)  (reg&= ~(1<<pos))                // dto. fuer loeschen
#define testbit(reg,pos) ((reg) & (1<<pos))               // testet an der Bitposition pos das Bit auf 1 oder 0



/* -------------------------------------------------------
     Prototypen
   ------------------------------------------------------- */

  void delay(char wert);
  void nibbleout(unsigned char wert, unsigned char hilo);
  void lcd_takt(void);
  void lcd_io(char wert);
  void lcd_init(void);
  void gotoxy(char x, char y);
  void lcd_putchar(char ch);
  void putramstring(char *c);
  void putromstring(const unsigned char *dataPtr);

  char wherex,wherey;

/* -------------------------------------------------------
     Implementationen
   ------------------------------------------------------- */
   
/* -------------------------------------------------------
      DELAY
      
      Zeitverzoegerung ist WERT * 0,1s
   ------------------------------------------------------- */

void delay(char wert)
{
   char i;
   for (i=0;i< wert;i++)
   {
     _delay_ms(100);
   }
}

/* -------------------------------------------------------
     NIBBLEOUT
     
     sendet ein Halbbyte an das LC-Display an die in
     den Defines angegebenen Pins
     
         WERT= gesamtes Byte
         HILO= 1 => oberen 4 Bits werden gesendet
         HILO= 0 => untere 4 Bits werden gesendet
   ------------------------------------------------------- */

void nibbleout(unsigned char wert, unsigned char hilo)
{
  if (hilo)
  {
     if (testbit(wert,7)) { setbit(lcd_db,lcd_d7); }
                    else  { clrbit(lcd_db,lcd_d7); }
     if (testbit(wert,6)) { setbit(lcd_db,lcd_d6); }
                    else  { clrbit(lcd_db,lcd_d6); }
     if (testbit(wert,5)) { setbit(lcd_db,lcd_d5); }
                    else  { clrbit(lcd_db,lcd_d5); }
     if (testbit(wert,4)) { setbit(lcd_db,lcd_d4); }
                    else  { clrbit(lcd_db,lcd_d4); }
  }
  else
  {
     if (testbit(wert,3)) { setbit(lcd_db,lcd_d7); }
                    else  { clrbit(lcd_db,lcd_d7); }
     if (testbit(wert,2)) { setbit(lcd_db,lcd_d6); }
                    else  { clrbit(lcd_db,lcd_d6); }
     if (testbit(wert,1)) { setbit(lcd_db,lcd_d5); }
                    else  { clrbit(lcd_db,lcd_d5); }
     if (testbit(wert,0)) { setbit(lcd_db,lcd_d4); }
                    else  { clrbit(lcd_db,lcd_d4); }
  }
}

/* -------------------------------------------------------
      LCD_TAKT
      
      gibt einen Clockimpuls an das Display
   ------------------------------------------------------- */

void lcd_takt(void)
{
  setbit(lcd_ctrl,lcd_e);
  _delay_us(60);
  clrbit(lcd_ctrl,lcd_e);
  _delay_us(60);
}

/* -------------------------------------------------------
      LCD_IO

      sendet ein Byte an das Display an die in den
      Defines angegebenen Pins
              Wert = zu sendendes Byte
   ------------------------------------------------------- */

void lcd_io(char wert)
{
  nibbleout(wert,1);
  lcd_takt();
  nibbleout(wert,0);
  lcd_takt();
}

/* -------------------------------------------------------
     LCD_INIT
     
     bereitet das Display fuer den Betrieb vor. Die in
     den Defines gegebenen Anschluesse des Displays 
     werden auf Ausgang geschaltet.
     Es wird 4-Bit Datenuebertragung verwendet.

   ------------------------------------------------------- */

void lcd_init(void)
{
  char i;

  lcd_dbio |= 1<<lcd_d7 | 1<<lcd_d6 | 1<<lcd_d5 | 1<<lcd_d4;
  lcd_ctrlio |= 1<<lcd_rs | 1<<lcd_e;

  clrbit(lcd_ctrl,lcd_rs);
  for (i= 0; i< 3; i++)
  {
     lcd_io(0x20);
	 _delay_ms(6);
  }
  lcd_io(0x28);
  _delay_ms(6);
  lcd_io(0x0c);
  _delay_ms(6);
  lcd_io(0x01);
  _delay_ms(6);
  wherex= 0; wherey= 0;
}

/* -------------------------------------------------------
     GOTOXY
     
     setzt den Textcursor an eine Stelle im Display. Die
     obere linke Ecke hat die Koordinate (1,1)
   ------------------------------------------------------- */

void gotoxy(char x, char y)
{
  unsigned char lcd_adr;

  lcd_adr=(0x80+((y-1)*0x40))+x-1;
  clrbit(lcd_ctrl,lcd_rs);
  lcd_io(lcd_adr);
  wherex= x;
  wherey= y;
}

/* -------------------------------------------------------
     LCD_PUTCHAR
     
     platziert ein Zeichen auf dem Display.
     
               CH = auszugebendes Zeichen
   ------------------------------------------------------- */

void lcd_putchar(char ch)
{
  setbit(lcd_ctrl,lcd_rs);
  lcd_io(ch);
  wherex++;
}

/* -------------------------------------------------------
      PUTRAMSTRING
      
      gibt einen AsciiZ Text der im RAM gespeichert ist
      auf dem Display aus.
      
      Bsp.:

      char strbuf[] = "H. Welt";

      putramstring(strbuf);

   ------------------------------------------------------- */

void putramstring(char *c)                              // Uebergabe eines Zeigers (auf einen String)
{
  while (*c)
  {
    lcd_putchar(*c++);
  }
}

/* -------------------------------------------------------
     PUTROMSTRING
     
     gibt einen Text, der im Flashrom gespeichert ist
     auf dem LC-Display aus (Wird vom Macro <prints>
     aufgerufen .

     Bsp.:

       static const uint8_t mytext[] PROGMEM = "Text";

       putromstring(PSTR("H. Welt"));
       putromstring(&mytext[0]);

       prints("Hallo");

   ------------------------------------------------------- */

void putromstring(const unsigned char *dataPtr)
{
  unsigned char c;

  for (c=pgm_read_byte(dataPtr); c; ++dataPtr, c=pgm_read_byte(dataPtr)) lcd_putchar(c);
}
