/* ----------------  TXLCD_TEMP.C    ----------------

     Thermometer mit Anzeige auf einem alpha-
     numerischen Textdisplay.

     Display:          Textdisplay mit HD44780 Controller
                       Betriebsmodus 4-Bit

                       Pollin Display C0802-04
                       (billiger gehts nicht)

     MCU:              ATtiny2313

     Quarz:            keiner / int. Takt 8 MHz

     Fuses             Lfuse 0xE4
                       Hfuse 0xDF
                       EFuse 0xFF

     Compiler:         AVR-GCC  4.7.2
     IDE:              Context

     uebersetzen mit:  GCCCOMPILE_GKZ.BAT


     21.06.2013        R. Seelig

   ------------------------------------------------------- */

/*
      Anschluss Pollin-Display C0802-04 an den Controller
      ---------------------------------------------------
         o +5V
         |                            Display                 Controller
         _                        Funktion   PIN            PIN    Funktion
        | |
        |_|                          GND      1 ------------
         |                          +5V       2 ------------
         o----o Kontrast   ---    Kontrast    3 ------------
        _|_                           RS      4 ------------  9      PD5
        \ /                          GND      5 ------------
        -'-                    (Takt) E       6 ------------  8      PD4
         |                           D4       7 ------------  2      PD0
        --- GND                      D5       8 ------------  3      PD1
                                     D6       9 ------------  6      PD2
                                     D7      10 ------------  7      PD3
                                                                                 I2C

                                                              19     ----------- SCL
                                                              17     ----------- SDA
*/
// ------------------------------------------------------------------------------------------------------------------
/*
      Anschluss NTC an den Controller
      ---------------------------------------------------

      siehe:  temp_tiny_pollin.gif
*/


#ifndef F_CPU
#define F_CPU 12000000
#endif


#include <util/delay.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <stdlib.h>

#include "tx_lcd.inc"

#define kommapos            2

#define chargedir           DDRD
#define chargeport          PORTD
#define chargepin           PD6                           // Pin, der den Kondensator laedt / entlaedt

#define chbit(wert,nr)      ((wert) & (1<<nr))            // testet auf einer Bitposition <nr> im Port <wert> auf 1 oder 0
#define compout             (chbit(ACSR,ACO))             // testet, ob Komperatoroutput 1 oder 0 ist

#define charge()            (chargeport |= (1<<chargepin))
#define discharge()         (chargeport &= ~(1<<chargepin))


// --------------------------------------------------------------------
//   Widerstandswerte des NTC 0.2 10k Widerstandssensor Fa. Reichelt
//   Widerstandwertsabstufung in 5C Schritten beginnend bei -25C
// --------------------------------------------------------------------

const int tempr[] = {1293,964,725,550,421,325,253,198,157,125,100,81,65,53,44,36,30,25,21,17,15,12,10,9,8,7};

// Tabelle wurde noetig, da Gleitkommazahlenberechnungen nach Steinhart-Hart nicht in den
// Speicher des ATtiny2313 passen


/*  --------------------------------------------------
                         MYFTOA

    Pseudo-Konverterfunktion. Der uebergebene Integer-
    wert wird so behandelt, als waere er eine
    "Kommazahl" mit Anzahl bestimmter Nachkomma-
    stellen die im Define "kommapos" angegeben ist.

    Bsp.:
      z= 12345 wird in einen String mit "123.45"

      konvertiert. "kommapos" bestimmt die
      Kommaposition von RECHTS gelesen. Angabe 3
      zeigt somit 2 Nachkommastellen an.

   --------------------------------------------------- */


void myftoa(int z, char *s)
{
  char l;
  char *p,*p2;                          // Hilfszeiger auf einen String
  char ch,i;

  l= 0;
  p=s;
  p2= s;
  itoa(z,s,10);                         // itoa verwenden, wenn Vorzeichen benoetigt wird
  while (*s)                            // Ende des Strings suchen
  {
    l++;                                // .. und somit die Lnge erfassen
    s++;
  }
  p=s;
  s--;
  if (l<kommapos)                       // wenn Dezimalstring kleiner 3-stellig ist
  {
    p= p+kommapos-l;
    *p= 0;                              // neues Ascii-Zero setzen
    p--;
    for (i= 0;i < kommapos;i++)         // mit fuehrenden Nullen auffuellen
    {
      ch= *s;
      if (s< p2) {ch= '0';}
      *p= ch;
      p--;
      s--;
    }
    l=kommapos;
  }
  p2 +=l+1;
  *p2= 0;
  p2--;
  *p2='y';
  s= p2;
  s--;
  for (i= 1; i< kommapos; i++)
  {
    ch= *s;
    *p2= ch;
    s--;
    p2--;
  }
  *p2= '.';
}

void printchars(char anz, char ch)
{
  char i;

  for (i= 0; i< anz; i++) {lcd_putchar(ch); }
}


void init_anacomp(void)
{
  DIDR= 0x03;                                 // Digitalinputpins fuer Komparatorpins abschalten (AIN0D und AIN1D = 1)
  ACSR &= ~(1<< ACD);                         // Enable Komparator

  chargedir |= (1<<chargepin);                // Anschluss, der den Kondensator laedt / entlaedt auf Ausgang setzen
}


uint16_t getmickeys(void)
{
  uint16_t    wertcnt;
  char        charge;

  discharge();
  delay(4);
  wertcnt= 0;
  charge();
  while ((!compout) && (wertcnt< 0xffff))
  {
    wertcnt++;
    _delay_us(3);                            // 3: ergibt 7650 Ticks bei 220kOhm / 220nF
                                             /* Bsp.Rechnung:
                                                         2200 / 7650 = 0,2875
                                                            2 / 7    = 0,2875
                                             */
  }
  discharge();
  return wertcnt;
}

/*  --------------------------------------------------
                         GETKOHM

      Ein von der Funktion <GETMICKEYS> gelieferter
      Wert gibt an einem 220k Widerstand einen Zaehler-
      stand von 5700 (100nF Kondensator) zurueck. Soll
      dieser Wert auf 2200 normiert werden, so ist der
      Faktor 2200/7650 = 0.3859. Das entspricht in
      etwa 2/7. Der Faktor 2/7 wird absichtlich mit
      Integerwerten generiert, damit die Gleitkomma-
      zahlenbibliothek (die sehr gross ist und nicht
      in einen ATtiny passt) nicht eingebunden werden
      muss! Der Wert 2200 entspricht hier dann 220kOhm
   --------------------------------------------------- */


uint16_t getkohm(uint16_t wert)
{
  wert= wert*2;
  wert= wert/7;
  return wert;
}

/*  --------------------------------------------------
                         GETTEMP

       Im Array <tempr> sind beginnend mit -25C
       in 5K Schritten die Widerstandswerte des NTC 
       hinterlegt. <GETTEMP> sucht die Tabelle ab,
       die Position des Eintrages bestimmt die 
       Temperatur. Liegt der uebergebene Wert (wahr-
       scheinlich) zwischen 2 Werten, so wird der 
       Wert linear interpoliert.
   --------------------------------------------------- */


int gettemp(int mwert)
{
  char cx;
  int tempw,w2,w3,diffw;

  cx= 0;
  tempw= -250;
  while (mwert< tempr[cx])                 // die "Suche" im Array, zwischen welchen Werten sich
  {                                        // der gemessene Wert befindet
    cx++;
    tempw= tempw+50;
  }
  if (mwert==tempr[cx]) {return tempw;}    // war es GENAU der Wert erfolgt keine Mittelung
  tempw= tempw - 50;
  diffw= tempr[cx-1]-tempr[cx];
  w2=  diffw-(mwert-tempr[cx]);
  w3= ((w2*50)/diffw)+tempw;
  return w3;
}

int main()
{

  char str16[16];

  int z;
  char b;

  lcd_init();
  init_anacomp();
  prints("Tiny-AVR");
  gotoxy(3,2); prints("2313");
  delay(10);
  gotoxy(1,1); prints("NTC-10k ");
  gotoxy(1,2); prints("        ");
  delay(10);
  gotoxy(1,1); prints("AVR-Temp");

  z= 0;
  while(1)
  {
    gotoxy(5,2); prints("      ");
    gotoxy(5,2);
    myftoa(z,str16);
    putramstring(str16);
    delay(3);
    z= getmickeys();             // mit 220nF ergibt es 7650 Mickeys fuer 220 kOhm
    z= getkohm(z);               // Mickes in einen Widerstand umrechnen
    z= gettemp(z);               // aus Look-Uptabelle die Temperatur ermitteln
  }
}
